unit ControlPoint;

interface

uses
  Classes, jpeg, ComCtrls, Windows, Graphics, Cmap;

const
  NVARS = 16;
  NXFORMS = 8;
  SUB_BATCH_SIZE = 10000;
  PREFILTER_WHITE = (1 shl 26);
  FILTER_CUTOFF = 2.5;

type
 PLongintArray = ^TLongintArray;
 TLongintArray = array[0..8192] of Longint;

type
  TPointsArray = array of array of double;
  TPointsiArray = array of array of Integer;

  TXForm = class
    vars: array [0..NVARS - 1] of double;   // normalized interp coefs between variations
    c: array [0..2,0..1] of double; // the coefs to the affine part of the function
    density: double;                // prob is this function is chosen. 0 - 1
    color: double;                  // color coord for this function. 0 - 1

    colori: integer;
    varsi: array [0..NVARS - 1] of Integer;
    ci00, ci01, ci10, ci11, ci20, ci21: integer;

    varType : integer;
    Constructor Create;
  end;

  TControlPoint = class
  private
    formvartype : integer;
  public
    xforms: array[0..NXFORMS-1] of TXForm;
    cmap: TColorMap;
    cmapindex: integer;
    time: double;
    brightness: double; // 1.0 = normal
    contrast: double; // 1.0 = normal
    gamma: double;
    Width: integer;
    Height: integer;
    spatial_oversample: integer;
    center: array[0..1] of double; // camera center
    vibrancy: double;              // blend between color algs (0=old,1=new)
    hue_rotation: double;          // applies to cmap, 0-1
    background: array[0..3] of double;
    zoom: double;                  // effects ppu and sample density
    pixels_per_unit: double;       // and scale
    spatial_filter_radius: double; // variance of gaussian
    sample_density: double;        // samples per pixel (not bucket)
    (* in order to motion blur more accurately we compute the logs of the
      sample density many times and average the results.  we interplate
      only this many times. *)
    nbatches: integer;              // this much color resolution.  but making it too high induces clipping
    white_level: integer;
    cmap_inter: integer;            // if this is true, then color map interpolates one entry
		                    // at a time with a bright edge
    pulse: array[0..1,0..1] of double;  // [i][0]=magnitute [i][1]=frequency */
    wiggle: array[0..1,0..1] of double; // frequency is /minute, assuming 30 frames/s */

    PropTable: array Of Integer;
    jpeg : TJPegImage;
  private
    procedure PreparePropTable;

  public
    procedure PrepareForInteger;

    procedure SaveToStringlist(sl: TStringlist);
    procedure SaveToFile(Filename: String);

    procedure ParseString(aString: string);
    procedure ParseStringList(sl: TStringlist);
    procedure RandomCP;
    procedure RandomCP1;
    procedure CalcBoundbox;

    class function interpolate(cp1,cp2: TControlPoint; Time: double): TControlPoint;
    procedure Iterate(NrPoints: integer; var Points: TPointsArray);
    procedure IterateI(NrPoints: integer; var Points: TPointsIArray);
    procedure IterateIType0(NrPoints: integer; var Points: TPointsIArray);

    procedure Render(Canvas: TCanvas; Progress: TProgressBar);

    function Clone: TControlPoint;

    constructor Create;
    destructor Destroy; override;

  end;

implementation

uses
  SysUtils, math;

{ TControlPoint }

constructor TControlPoint.Create;
var
  i: Integer;
begin
  for i := 0 to NXFORMS -1 do begin
    XForms[i] := TXForm.Create;
  end;

  pulse[0][0] := 0;
  pulse[0][1] := 60;
  pulse[1][0] := 0;
  pulse[1][1] := 60;

  wiggle[0][0] := 0;
  wiggle[0][1] := 60;
  wiggle[1][0] := 0;
  wiggle[1][1] := 60;

  background[0] := 0;
  background[1] := 0;
  background[2] := 0;

  center[0] := 0;
  center[1] := 0;

  pixels_per_unit := 50;

  width := 100;
  Height := 100;

  spatial_oversample := 1;
  spatial_filter_radius := 0.5;

  gamma := 1;
  vibrancy := 1;
  contrast := 1;
  brightness := 1;

  sample_density := 50;
  zoom := 0;
  nbatches := 1;

  white_level := 200;
end;

destructor TControlPoint.Destroy;
var
  i: Integer;
begin
  for i := 0 to NXFORMS -1 do begin
    XForms[i].Free;
  end;

  inherited;
end;

procedure TControlPoint.PreparePropTable;
var
  i: Integer;
  propsum: double;
  LoopValue: double;
  j: integer;
  TotValue: double;
begin
  SetLength(PropTable, 1024);

  totValue := 0;
  for i := 0 to NXFORMS - 1 do begin
    totValue := totValue + Xforms[i].density;
  end;

  LoopValue := 0;
  for i := 0 to 1023 do begin
    propsum := 0;
    j := -1;
    repeat
      inc(j);
      propsum := propsum + Xforms[j].density;
    until (propsum > LoopValue) or (j = NXFORMS - 1);
    PropTable[i] := j;
    LoopValue := LoopValue + TotValue/1024;
  end;
end;

procedure TControlPoint.Iterate(NrPoints: integer;
  var Points: TPointsArray);
var
  i: Integer;
  px,py,pc: double;
  tx,ty: double;
  nx,ny: double;
  r: double;
  a: double;
begin
  px := 2 * random - 1;
  py := 2 * random - 1;
  pc := random;

  PreparePropTable;

  for i := -100 to NrPoints - 1 do begin
    With xforms[PropTable[Random(1024)]] do begin
      // first compute the color coord
      pc := (pc + color) / 2.0;

      try
        // then apply the affine part of the function
        tx := c[0][0] * px + c[1][0] * py + c[2][0];
        ty := c[0][1] * px + c[1][1] * py + c[2][1];

        px := 0;
        py := 0;

        // then add in proportional amounts of each of the variations
        if vars[0] > 0 then begin // linear
          px := px + vars[0] * tx;
          py := py + vars[0] * ty;
        end;
        if vars[1] > 0 then begin // sinusoidal
          px := px + vars[1] * sin(tx);
          py := py + vars[1] * cos(ty);
        end;
        if vars[2] > 0 then begin // complex
          r := tx * tx + ty * ty + 1E-6;
          px := px + vars[2] * tx/r;
          py := py + vars[2] * ty/r;
        end;
        if vars[3] > 0 then begin // swirl
          r := tx * tx + ty * ty;
          px := px + vars[3] * (sin(r) * tx - cos(r) * ty);
          py := py + vars[3] * (cos(r) * tx + sin(r) * ty);
        end;
        if vars[4] > 0 then begin // swirl
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;

          px := px + vars[4] * (sin(a) * tx - cos(a) * ty);
          py := py + vars[4] * (cos(a) * tx + sin(a) * ty);
        end;
        if vars[5] > 0 then begin // no name
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)/ PI
          else
            a := 0;

          r := sqrt(tx * tx + ty * ty) - 1;

          px := px + vars[5] * a;
          py := py + vars[5] * r;
        end;

        if vars[6] > 0 then begin // bent
          nx := tx;
          ny := ty;
          if (nx < 0) and (nx > -1E100) then  nx := nx * 2;
          if ny < 0 then  ny := ny / 2;

          px := px + vars[6] * nx;
          py := py + vars[6] * ny;
        end;

        if vars[7] > 0 then begin // Hart shaped box
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[7] * (sin(a * r) * r);
          py := py - vars[7] * (cos(a * r) * r);
        end;

        if vars[8] > 0 then begin // The world in a sphere
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty) ;


          px := px + vars[8] * (sin(r) * (a));
          py := py + vars[8] * (cos(r) * (a));
        end;

        if vars[9] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := power(tx * tx + ty * ty, 0.5) + 1E-6;

          px := px + vars[9] * ((cos(a) + sin(r)) / r);
          py := py + vars[9] * ((sin(a) - cos(r)) / r);
        end;

        if vars[10] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := power(tx * tx + ty * ty, 0.25) + 1E-6;

          px := px + vars[10] * (sin(a) /r);
          py := py - vars[10] * (cos(a) * r);
        end;

        if vars[11] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty) + 1E-6;

          px := px + vars[11] * ( sin(a) /r);
          py := py - vars[11] * ( cos(a) /r);
        end;

        if vars[12] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[12] * ( sin(a ) * cos(r) );
          py := py - vars[12] * ( cos(a ) * sin(r));
        end;

        if vars[13] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[13] * (sign(sin(a + r)) * sqr(sin(a + r)) * r);
          py := py - vars[13] * (sign(cos(a - r)) * sqr(cos(a - r)) * r);
        end;

        if vars[14] > 0 then begin // Square
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[14] * (sign(sin(a + r)) * sqr(sin(a + r)) * r);
          py := py - vars[14] * (sign(cos(a + r)) * sqr(cos(a + r)) * r);
        end;

        if vars[15] > 0 then begin // Folded hankercief
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[15] * (sin(a + r) * r);
          py := py - vars[15] * (cos(a - r) * r);
        end;
      except
        on EMathError do begin
//          raise Exception.Create('Iteration blows up');
          exit;
        end;
      end;
    end;

    // store points
    if i >= 0 then begin
      Points[i][0] := px;
      Points[i][1] := py;
      Points[i][2] := pc;
    end
  end;
end;

procedure TControlPoint.ParseString(aString: string);
var
  ParseValues: TStringList;
  ParsePos: integer;
  CurrentToken: string;
  CurrentXForm: integer;
  i: integer;
  totvar: double;
  OldDecimalSperator: Char;
begin
  ParseValues := TStringList.Create;
  ParseValues.CommaText := AString;

  OldDecimalSperator := DecimalSeparator;
  DecimalSeparator := '.';

  CurrentXForm := 0;

  ParsePos := 0;
  while (ParsePos < ParseValues.Count) do begin
    CurrentToken := ParseValues[ParsePos];
    if AnsiCompareText(CurrentToken, 'xform') = 0 then begin
      Inc(ParsePos);
      CurrentXForm := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'time') = 0 then begin
      Inc(ParsePos);
      time := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'brightness') = 0 then begin
      Inc(ParsePos);
      brightness := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'contrast') = 0 then begin
      Inc(ParsePos);
      contrast := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'gamma') = 0 then begin
      Inc(ParsePos);
      gamma := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'vibrancy') = 0 then begin
      Inc(ParsePos);
      vibrancy := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'hue_rotation') = 0 then begin
      Inc(ParsePos);
      hue_rotation := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'zoom') = 0 then begin
      Inc(ParsePos);
      zoom := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'imagesize') = 0 then begin
      Inc(ParsePos);
      Width  := StrToInt(ParseValues[ParsePos]);
      Inc(ParsePos);
      Height := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'image_size') = 0 then begin
      Inc(ParsePos);
      Width  := StrToInt(ParseValues[ParsePos]);
      Inc(ParsePos);
      Height := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'center') = 0 then begin
      Inc(ParsePos);
      center[0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      center[1] := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'background') = 0 then begin
      Inc(ParsePos);
      background[0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      background[1] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      background[2] := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'pulse') = 0 then begin
      Inc(ParsePos);
      pulse[0,0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      pulse[0,1] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      pulse[1,0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      pulse[1,1] := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'wiggle') = 0 then begin
      Inc(ParsePos);
      wiggle[0,0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      wiggle[0,1] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      wiggle[1,0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      wiggle[1,1] := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'pixels_per_unit') = 0 then begin
      Inc(ParsePos);
      pixels_per_unit := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'spatial_filter_radius') = 0 then begin
      Inc(ParsePos);
      spatial_filter_radius := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'spatial_oversample') = 0 then begin
      Inc(ParsePos);
      spatial_oversample := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'sample_density') = 0 then begin
      Inc(ParsePos);
      sample_density := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'nbatches') = 0 then begin
      Inc(ParsePos);
      nbatches := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'white_level') = 0 then begin
      Inc(ParsePos);
      white_level := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'cmap') = 0 then begin
      Inc(ParsePos);
      cmapindex := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'cmap_inter') = 0 then begin
      Inc(ParsePos);
      cmap_inter := StrToInt(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'palette') = 0 then begin
//      Inc(ParsePos);
//      cmapindex := StrToInt(ParseValues[ParsePos]);
      OutputDebugString(Pchar('NYI import Palette'));
    end else if AnsiCompareText(CurrentToken, 'density') = 0 then begin
      Inc(ParsePos);
      xforms[CurrentXForm].Density := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'color') = 0 then begin
      Inc(ParsePos);
      xforms[CurrentXForm].color := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'coefs') = 0 then begin
      Inc(ParsePos);
      xforms[CurrentXForm].c[0,0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      xforms[CurrentXForm].c[0,1] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      xforms[CurrentXForm].c[1,0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      xforms[CurrentXForm].c[1,1] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      xforms[CurrentXForm].c[2,0] := StrToFloat(ParseValues[ParsePos]);
      Inc(ParsePos);
      xforms[CurrentXForm].c[2,1] := StrToFloat(ParseValues[ParsePos]);
    end else if AnsiCompareText(CurrentToken, 'var') = 0 then begin
      for i := 0 to NVARS - 1 do begin
        xforms[CurrentXForm].vars[i] := 0;
      end;

      i := 0;
      while true do begin
        if (ParsePos + 1) >=  ParseValues.Count then
          break;
        if ParseValues[ParsePos + 1][1] in ['a'..'z','A'..'Z'] then
          break;

        Inc(ParsePos);
        xforms[CurrentXForm].vars[i] := StrToFloat(ParseValues[ParsePos]);
        Inc(i);
      end;

      totvar := 0;
      for i := 0 to NVARS - 1 do begin
        totvar := totvar + xforms[CurrentXForm].vars[i];
      end;
      for i := 0 to NVARS - 1 do begin
        xforms[CurrentXForm].vars[i] := xforms[CurrentXForm].vars[i]/totvar;
      end;
    end else begin
      OutputDebugString(Pchar('Unknown Token: ' + CurrentToken));
    end;

    Inc(ParsePos);
  end;
  GetCmap(cmapindex, hue_rotation, Cmap);

  ParseValues.Free;

  DecimalSeparator := OldDecimalSperator;
end;

procedure TControlPoint.Render(Canvas: TCanvas; Progress: TProgressBar);
var
  image_Width: integer;
  image_Height: integer;
  BucketWidth: integer;
  BucketHeight: integer;
  gutter_width : integer;

  buckets: array of array of integer;
  accumulate: array of array of integer;

  cmap: array[0..255,0..3] of integer;

  i,j,k: integer;

  scale : double;
  locsample_density: double;
  t0,t1: double;
  corner0,corner1: double;
  bounds: array[0..3] of double;
  size: array[0..1] of double;
  shift: integer;
  nsamples: integer;
  ppux, ppuy: double;

  nrbatches: integer;

  points: TPointsArray;

  color_index : integer;
  bx,by: integer;

  k1,k2: double;
  area: double;
  bg: array[0..2] of double;
  vib_gam_n: integer;
  loc_vibrancy: double;
  loc_gamma: double;
  ls: double;

  bm: TBitmap;

  alpha: double;
  r,g,b: double;

  oversample: integer;
  filter_width: integer;
  filter: array of array of double;
  t: double;

  ii,jj: integer;
  fp: array[0..3] of double;
begin
  image_height := Height;
  image_Width := Width;

  // calc filter
  oversample := spatial_oversample;
  filter_width := Round(2.0 * FILTER_CUTOFF * oversample * spatial_filter_radius);
  // make sure it has same parity as oversample
  if odd(filter_width + oversample) then
    inc(filter_width);

  setLength(filter, filter_width, filter_width);
  for i := 0 to filter_width- 1 do begin
    for j := 0 to filter_width- 1 do begin
      filter[i,j] := exp(-2.0 * power(((2.0 * i + 1.0) / filter_width - 1.0) * FILTER_CUTOFF, 2) *
                                power(((2.0 * j + 1.0) / filter_width - 1.0) * FILTER_CUTOFF, 2));
    end;
  end;
  //normalize filter
  t := 0;
  for i := 0 to filter_width- 1 do begin
    for j := 0 to filter_width- 1 do begin
      t := t + filter[i,j];
    end;
  end;
  for i := 0 to filter_width- 1 do begin
    for j := 0 to filter_width- 1 do begin
      filter[i,j] := filter[i,j]/t;
    end;
  end;

  // allocate memory
  gutter_width := (filter_width - oversample) div 2;
  BucketHeight := oversample * image_height + 2 * gutter_width;
  Bucketwidth  := oversample * image_width  + 2 * gutter_width;

  SetLength(buckets, BucketWidth * BucketHeight, 4);
  SetLength(accumulate, BucketWidth * BucketHeight, 4);

  // Clear memory
  for i := 0 to BucketWidth * BucketHeight -1 do begin
    buckets[i][0] := 0;
    buckets[i][1] := 0;
    buckets[i][2] := 0;
    buckets[i][3] := 0;

    accumulate[i][0] := 0;
    accumulate[i][1] := 0;
    accumulate[i][2] := 0;
    accumulate[i][3] := 0;
  end;

  // create color map
  for i := 0 to 255 do begin
    for j := 0 to 2 do begin
      cmap[i][j] := self.cmap[i][j];
    end;
    cmap[i][3] := white_level;
  end;

  // camera
  scale := power(2, zoom);
  locsample_density := sample_density * scale * scale;
  ppux := pixels_per_unit * scale;
  ppuy := pixels_per_unit * scale;
  // todo field stuff
  shift := 0;
  t0 := gutter_width / (oversample * ppux);
  t1 := gutter_width / (oversample * ppuy);
  corner0 := center[0] - image_width / ppux / 2.0;
  corner1 := center[1] - image_height / ppuy / 2.0;
  bounds[0] := corner0 - t0;
  bounds[1] := corner1 - t1 + shift;
  bounds[2] := corner0 + image_width  / ppux + t0;
  bounds[3] := corner1 + image_height / ppuy + t1 + shift;
  size[0] := 1.0 / (bounds[2] - bounds[0]);
  size[1] := 1.0 / (bounds[3] - bounds[1]);

  nsamples := Round(sample_density * bucketwidth * BucketHeight / (oversample * oversample));

  nrbatches := round(nsamples / (nbatches * SUB_BATCH_SIZE));
  Progress.Max := nrbatches - 1;
  for i := 0 to nrbatches do begin
    Progress.Position := i;
    // generate points
    iterate(SUB_BATCH_SIZE, points);

    // add points to buckets
    for j := 0 to SUB_BATCH_SIZE - 1 do begin
      if ((points[j][0] < bounds[0]) or (points[j][0] > bounds[2]) or
          (points[j][1] < bounds[1]) or (points[j][1] > bounds[3])) then
        continue;

      color_index := round(points[j][2] * 256) mod 256;
      if color_index < 0 then
        color_index := 0;

      bx := floor(BucketWidth * (points[j][0] - bounds[0]) * size[0]);
      by := floor(BucketHeight * (points[j][1] - bounds[1]) * size[1]);
      for k := 0 to 3 do begin
        buckets[bx + by * BucketWidth][k] := buckets[bx + by * BucketWidth][k] + cmap[color_index][k];
      end;
    end;
  end;

  // fill acculumation buffer
  loc_vibrancy := vibrancy;
  loc_gamma := gamma;
  vib_gam_n := 1;
  bg[0] := background[0];
  bg[1] := background[1];
  bg[2] := background[2];

  k1 := (Contrast * brightness * 268 * PREFILTER_WHITE) / 256;
  area := image_width * image_height / (ppux * ppuy);
  k2 := (oversample * oversample * nbatches)/(Contrast * area * White_level * sample_density);

  for i := 0 to BucketWidth * BucketHeight -1 do begin
    if Buckets[i][3] = 0 then
      Continue;

    ls := 0;
    try
      ls := (k1 * log10(1 + Buckets[i][3] * k2))/Buckets[i][3];
    except
    end;
    accumulate[i,0] := accumulate[i,0] + Round(0.5 + Buckets[i,0] * ls);
    accumulate[i,1] := accumulate[i,1] + Round(0.5 + Buckets[i,1] * ls);
    accumulate[i,2] := accumulate[i,2] + Round(0.5 + Buckets[i,2] * ls);
    accumulate[i,3] := accumulate[i,3] + Round(0.5 + Buckets[i,3] * ls);
  end;

  // filter the accumulation buffer down into the image
  bm := TBitmap.Create;
  bm.PixelFormat := pf32bit;
  bm.Width := image_Width;
  bm.Height := image_Height;


  loc_vibrancy := loc_vibrancy/vib_gam_n;
  loc_gamma := vib_gam_n/loc_gamma;

  bg[0] := 256 * bg[0]/vib_gam_n;
  bg[1] := 256 * bg[1]/vib_gam_n;
  bg[2] := 256 * bg[2]/vib_gam_n;

  Progress.Max := Image_Height-1;
  for i := 0 to Image_Height - 1 do begin
    Progress.Position := i;
    for j := 0 to Image_Width - 1 do begin
      // todo filter

      fp[0] := 0;
      fp[1] := 0;
      fp[2] := 0;
      fp[3] := 0;
      for ii := 0 to filter_width- 1 do begin
        for jj := 0 to filter_width- 1 do begin
          fp[0] := fp[0] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][0];
          fp[1] := fp[1] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][1];
          fp[2] := fp[2] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][2];
          fp[3] := fp[3] + filter[ii,jj] * accumulate[(i * oversample + ii) * BucketWidth + j * oversample  + jj][3];
        end;
      end;

      alpha := fp[3];
      if (alpha > 0.0) then begin
        ls := loc_vibrancy * 256.0 * power(alpha / PREFILTER_WHITE, loc_gamma)/ (alpha / PREFILTER_WHITE);
        alpha := power(alpha / PREFILTER_WHITE, loc_gamma);
	if (alpha < 0.0) then
          alpha := 0.0
        else if (alpha > 1.0) then
          alpha := 1.0;
      end else begin
        ls := 0;
      end;

      r := ls * fp[0] / PREFILTER_WHITE;
      if (loc_vibrancy < 1.0) then
	r := r + (1.0 - loc_vibrancy) * 256.0 * power(fp[0] / PREFILTER_WHITE, loc_gamma);
      r := r + ((1.0 - alpha) * bg[0]);
      if (r < 0) then
         r := 0
      else if (r > 255) then
        r := 255;

      g := ls * fp[1] / PREFILTER_WHITE;
      if (loc_vibrancy < 1.0) then
	g := g + (1.0 - loc_vibrancy) * 256.0 * power(fp[1] / PREFILTER_WHITE, loc_gamma);
      g := g + ((1.0 - alpha) * bg[1]);
      if (g < 0) then
         g := 0
      else if (g > 255) then
        g := 255;

      b := ls * fp[2] / PREFILTER_WHITE;
      if (loc_vibrancy < 1.0) then
	b := b + (1.0 - loc_vibrancy) * 256.0 * power(fp[2] / PREFILTER_WHITE, loc_gamma);
      b := b + ((1.0 - alpha) * bg[2]);
      if (g < 0) then
         b := 0
      else if (b > 255) then
        b := 255;

      PLongintArray(bm.scanline[i])[j] := RGB(round(b), round(g), round(r));

    end;
  end;

  Canvas.Draw(0,0, bm);

  if assigned(jpeg) then
    jpeg.free;

  jpeg := TJPegImage.Create;
  jpeg.assign(bm);

  bm.Free;
end;

procedure TControlPoint.RandomCP;
const
  xform_distrib: array [0..12] of integer = (2,2,2,3,3,3,4,4,5,5,6,7,8);
  var_distrib: array [0..36] of integer = (-1,-1,-1,-1,-1,-1,-1,-1,-1,0,0,0,0,1,1,1,2,2,2,3,3,4,4,5,5,6,6,7,7,8,9,10,11,12,13,14,15);
  mixed_var_distrib: array [0..26] of integer = (0,0,0,1,1,1,2,2,2,3,3,4,4,5,5,6,6,7,7,8,9,10,11,12,13,14,15);
var
  nrXforms: integer;
  i,j: integer;
  rv: integer;
begin
  hue_rotation := random;
  cmapindex := RANDOMCMAP;
  GetCmap(cmapindex, hue_rotation, cmap);
  time := 0.0;
  nrXforms := xform_distrib[random(13)];

  rv := var_distrib[random(37)];

  for i := 0 to NXFORMS - 1 do begin
    xforms[i].density := 0;
  end;

  for i:= 0 to nrXforms - 1 do begin
    xforms[i].density := 1.0/nrXforms;
    if i = 0 then
      xforms[i].color := 1
    else
      xforms[i].color := 0;

    xforms[i].c[0][0] := 2 * random - 1;
    xforms[i].c[0][1] := 2 * random - 1;
    xforms[i].c[1][0] := 2 * random - 1;
    xforms[i].c[1][1] := 2 * random - 1;
    xforms[i].c[2][0] := 4 * random - 2;
    xforms[i].c[2][1] := 4 * random - 2;

    for j := 0 to NVARS - 1 do begin
      xforms[i].vars[j] := 0;
    end;

    if rv < 0 then
      xforms[i].vars[Mixed_var_distrib[random(27)]] := 1
    else
      xforms[i].vars[rv] := 1;
  end;

  CalcBoundbox;
end;

procedure TControlPoint.RandomCP1;
var
  i,j: Integer;
begin
  RandomCP;

  for i:= 0 to NXFORMS - 1 do begin
    for j := 0 to NVARS - 1 do begin
      xforms[i].vars[j] := 0;
    end;
      xforms[i].vars[0] := 1;
  end;

  CalcBoundbox;
end;

procedure TControlPoint.CalcBoundbox;
var
  Points: TPointsArray;
  i,j: integer;
  deltax, minx, maxx: double;
  cntminx, cntmaxx: integer;
  deltay, miny, maxy: double;
  cntminy, cntmaxy: integer;
  LimitOutSidePoints: integer;
begin
  SetLength(Points,SUB_BATCH_SIZE,3);
  iterate(SUB_BATCH_SIZE, Points);

  LimitOutSidePoints := Round(0.05 * SUB_BATCH_SIZE);

  minx := 1e10;
  maxx := -1e10;
  miny := 1e10;
  maxy := -1e10;
  for i := 0 to SUB_BATCH_SIZE - 1 do begin
    minx := min(minx,Points[i][0]);
    maxx := max(maxx,Points[i][0]);
    miny := min(miny,Points[i][1]);
    maxy := max(maxy,Points[i][1]);
  end;

  deltax := (maxx - minx) * 0.25;
  maxx := (maxx + minx)/2;
  minx := maxx;

  deltay := (maxy - miny) * 0.25;
  maxy := (maxy + miny)/2;
  miny := maxy;

  for j:= 0 to 10 do begin
    cntminx := 0;
    cntmaxx := 0;
    cntminy := 0;
    cntmaxy := 0;
    for i := 0 to SUB_BATCH_SIZE - 1 do begin
      if (Points[i][0] < minx) then Inc(cntminx);
      if (Points[i][0] > maxx) then Inc(cntmaxx);
      if (Points[i][1] < miny) then Inc(cntminy);
      if (Points[i][1] > maxy) then Inc(cntmaxy);
    end;

    if (cntMinx < LimitOutSidePoints) then begin
      minx := minx + deltax;
    end else begin
      minx := minx - deltax;
    end;

    if (cntMaxx < LimitOutSidePoints) then begin
      maxx := maxx - deltax;
    end else begin
      maxx := maxx + deltax;
    end;

    deltax := deltax / 2;

    if (cntMiny < LimitOutSidePoints) then begin
      miny := miny + deltay;
    end else begin
      miny := miny - deltay;
    end;

    if (cntMaxy < LimitOutSidePoints) then begin
      maxy := maxy - deltay;
    end else begin
      maxy := maxy + deltay;
    end;

    deltay := deltay / 2;
  end;

  center[0] := (minx + maxx)/2;
  center[1] := (miny + maxy)/2;

  if ((maxx - minx) > 0.001) and ((maxy - miny) > 0.001) then
//    pixels_per_unit := 0.5 * Min(width/(maxx - minx), Height/(maxy - miny));
    pixels_per_unit := 0.7 * Min(width/(maxx - minx), Height/(maxy - miny))
  else
    pixels_per_unit := 10;
end;

class function TControlPoint.interpolate(cp1, cp2: TControlPoint; Time: double): TControlPoint;
var
  c0, c1: double;
  i,j: integer;
  r,s,t: array[0..2] of double;
  totvar: double;
  {z,rhtime: double;}
begin
  if (cp2.time - cp1.time) > 1e-6 then begin
    c0 := (cp2.time - time) / (cp2.time - cp1.time);
    c1 := 1 - c0;
  end else begin
    c0 := 1;
    c1 := 0;
  end;

  Result := TControlPoint.Create;
  Result.time := Time;

  if cp1.cmap_inter = 0 then
    for i := 0 to 255 do begin
      r[0] := cp1.cmap[i][0]/255;
      r[1] := cp1.cmap[i][1]/255;
      r[2] := cp1.cmap[i][2]/255;
      rgb2hsv(r, s);
      r[0] := cp2.cmap[i][0]/255;
      r[1] := cp2.cmap[i][1]/255;
      r[2] := cp2.cmap[i][2]/255;
      rgb2hsv(r, t);
      t[0] := c0 * s[0] + c1 * t[0];
      t[1] := c0 * s[1] + c1 * t[1];
      t[2] := c0 * s[2] + c1 * t[2];
      hsv2rgb(t, r);
      Result.cmap[i][0] := Round(255 * r[0]);
      Result.cmap[i][1] := Round(255 * r[1]);
      Result.cmap[i][2] := Round(255 * r[2]);
    end;

  Result.cmapindex := -1;

  Result.brightness := c0 * cp1.brightness + c1 * cp2.brightness;
  Result.contrast := c0 * cp1.contrast + c1 * cp2.contrast;
  Result.gamma := c0 * cp1.gamma + c1 * cp2.gamma;
  Result.vibrancy := c0 * cp1.vibrancy + c1 * cp2.vibrancy;
  Result.width := cp1.width;
  Result.height := cp1.height;
  Result.spatial_oversample := Round(c0 * cp1.spatial_oversample + c1 * cp2.spatial_oversample);
  Result.center[0] := c0 * cp1.center[0] + c1 * cp2.center[0];
  Result.center[1] := c0 * cp1.center[1] + c1 * cp2.center[1];
  Result.pixels_per_unit := c0 * cp1.pixels_per_unit + c1 * cp2.pixels_per_unit;
  Result.background[0] := c0 * cp1.background[0] + c1 * cp2.background[0];
  Result.background[1] := c0 * cp1.background[1] + c1 * cp2.background[1];
  Result.background[2] := c0 * cp1.background[2] + c1 * cp2.background[2];
  Result.spatial_filter_radius := c0 * cp1.spatial_filter_radius + c1 * cp2.spatial_filter_radius;
  Result.sample_density := c0 * cp1.sample_density + c1 * cp2.sample_density;
  Result.zoom := c0 * cp1.zoom + c1 * cp2.zoom;
  Result.nbatches := Round(c0 * cp1.nbatches + c1 * cp2.nbatches);
  Result.white_level := Round(c0 * cp1.white_level + c1 * cp2.white_level);

  for i := 0 to 3 do begin
    Result.pulse[i div 2][i mod 2] := c0 * cp1.pulse[i div 2][i mod 2] + c1 * cp2.pulse[i div 2][i mod 2];
    Result.wiggle[i div 2][i mod 2] := c0 * cp1.wiggle[i div 2][i mod 2] + c1 * cp2.wiggle[i div 2][i mod 2];
  end;

  for i:= 0 to NXFORMS - 1 do begin
    Result.xforms[i].density := c0 * cp1.xforms[i].density + c1 * cp2.xforms[i].density;
    Result.xforms[i].color := c0 * cp1.xforms[i].color + c1 * cp2.xforms[i].color;
    for j := 0 to NVARS - 1 do begin
      Result.xforms[i].vars[j] := c0 * cp1.xforms[i].vars[j] + c1 * cp2.xforms[i].vars[j];
    end;

    totvar := 0;
    for j := 0 to NVARS - 1 do begin
      totvar := totvar + Result.xforms[i].vars[j];
    end;
    for j := 0 to NVARS - 1 do begin
      Result.xforms[i].vars[j] := Result.xforms[i].vars[j]/totvar;
    end;

    // interpol matrix
    for j := 0 to 2 do begin
      Result.xforms[i].c[j,0] := c0 * cp1.xforms[i].c[j,0] + c1 * cp2.xforms[i].c[j,0];
      Result.xforms[i].c[j,1] := c0 * cp1.xforms[i].c[j,1] + c1 * cp2.xforms[i].c[j,1];
    end;

{
    rhtime := time * 2 * PI / (60.0 * 30.0);
    // pulse
    z := 1;
    for j := 0 to 1 do begin
      z := z + Result.pulse[j, 0] * sin(Result.pulse[j, 1] * rhtime)
    end;

    for j := 0 to 2 do begin
      Result.xforms[i].c[j][0] := Result.xforms[i].c[j][0] * z;
      Result.xforms[i].c[j][1] := Result.xforms[i].c[j][1] * z;
    end;

    // wiggle
    for j := 0 to 1 do begin
      z := Result.wiggle[j,1] * rhtime;

      Result.xforms[i].c[0][0] := Result.xforms[i].c[0][0] + Result.wiggle[j,0] * cos(z);
      Result.xforms[i].c[1][0] := Result.xforms[i].c[1][0] + Result.wiggle[j,0] * -sin(z);
      Result.xforms[i].c[0][1] := Result.xforms[i].c[0][1] + Result.wiggle[j,0] * sin(z);
      Result.xforms[i].c[1][1] := Result.xforms[i].c[1][1] + Result.wiggle[j,0] * cos(z);
    end;
}
  end;
end;

procedure TControlPoint.SaveToFile(Filename: String);
var
  sl: TStringlist;
begin
  sl := TStringlist.Create;

  SaveToStringlist(sl);

  sl.SaveToFile(filename);
  sl.Free;
end;

procedure TControlPoint.SaveToStringlist(sl: TStringlist);
var
  i,j: Integer;
  s: string;
  OldDecimalSperator: Char;
begin
  OldDecimalSperator := DecimalSeparator;
  DecimalSeparator := '.';

  sl.add(format('time %f', [time]));
  if cmapindex >= 0 then
    sl.add(format('cmap %d', [cmapindex]));
  sl.add(format('image_size %d %d center %f %f pixels_per_unit %f',
                [Width, Height, center[0], center[1], pixels_per_unit]));
  sl.add(format('spatial_oversample %d spatial_filter_radius %f',
                [spatial_oversample, spatial_filter_radius]));
  sl.add(format('sample_density %f', [sample_density]));
  sl.add(format('nbatches %d white_level %d background %f %f %f',
                [nbatches, white_level, background[0], background[1], background[2]]));
  sl.add(format('brightness %f gamma %f vibrancy %f hue_rotation %f cmap_inter %d',
                 [brightness, gamma, vibrancy, hue_rotation, cmap_inter]));

  for i:= 0 to NXFORMS - 1 do begin
    if xforms[i].density = 0 then
      Continue;

    sl.add(format('xform %d density %f color %f', [i, xforms[i].density, xforms[i].color]));
    s := 'var';
    for j := 0 to NVARS - 1 do begin
      s := format('%s %f',[s, xforms[i].vars[j]]);
    end;
    sl.add(s);
    sl.Add(format('coefs %f %f %f %f %f %f',
                  [xforms[i].c[0][0], xforms[i].c[0][1],
                   xforms[i].c[1][0], xforms[i].c[1][1],
                   xforms[i].c[2][0], xforms[i].c[2][1]]));
  end;

  DecimalSeparator := OldDecimalSperator;
end;

procedure TControlPoint.IterateIType0(NrPoints: integer;
  var Points: TPointsIArray);
var
  pix,piy,pic: Integer;
  i: integer;
begin
  pix := round((2 * random - 1) * 1024);
  piy := round((2 * random - 1) * 1024);
  pic := round(random * 256 * 1024);

  for i := -100 to NrPoints - 1 do begin
    With xforms[PropTable[Random(1024)]] do begin
      // first compute the color coord
      pic := (pic + colori) div 2;

      try
        if varsi[0] = 1024 then begin
          pix := (ci00 * pix + ci10 * piy + ci20) div 1024;
          piy := (ci01 * pix + ci11 * piy + ci21) div 1024;
        end else begin
          pix := (varsi[0] * ((ci00 * pix + ci10 * piy + ci20) div 1024)) div 1024;
          piy := (varsi[0] * ((ci01 * pix + ci11 * piy + ci21) div 1024)) div 1024;
        end;
      except
        on EMathError do begin
//          raise Exception.Create('Iteration blows up');
          exit;
        end;
      end;
    end;

    // store points
    if i >= 0 then begin
      Points[i][0] := pix;
      Points[i][1] := piy;
      Points[i][2] := pic div 1024;
    end
  end;

end;

procedure TControlPoint.IterateI(NrPoints: integer;
  var Points: TPointsIArray);
var
  i: Integer;
  px,py,pc: double;
  {pix,piy,}pic: Integer;
  tx,ty: double;
//  tix,tiy: Integer;
  nx,ny: double;
  r: double;
//  ri: integer;
  a: double;
begin
  if formvartype = 0 then begin
    IterateIType0(NrPoints, Points);
    exit;
  end;

  px := 2 * random - 1;
  py := 2 * random - 1;
  pc := random;
//  pix := round(px * 1024);
//  piy := round(py * 1024);
  pic := round(pc * 256 * 1024);

  for i := -100 to NrPoints - 1 do begin
    With xforms[PropTable[Random(1024)]] do begin
      // first compute the color coord
      pic := (pic + colori) div 2;

      try
        // then apply the affine part of the function
//        tix := (ci00 * pix + ci10 * piy + ci20) div 1024;
//        tiy := (ci01 * pix + ci11 * piy + ci21) div 1024;
        tx := c[0][0] * px + c[1][0] * py + c[2][0];
        ty := c[0][1] * px + c[1][1] * py + c[2][1];
//        tx := tix/1024.0;
//        ty := tiy/1024.0;

//        pix := 0;
//        piy := 0;
        px := 0;
        py := 0;

        // then add in proportional amounts of each of the variations
        if varsi[0] > 0 then begin // linear
          px := px + vars[0] * tx;
          py := py + vars[0] * ty;

//          pix := pix + (varsi[0] * tix) div 1024;
//          piy := piy + (varsi[0] * tiy) div 1024;
//          px := px + pix/1024.0;
//          py := py + piy/1024.0;
        end;
        if varsi[1] > 0 then begin // sinusoidal
          px := px + vars[1] * sin(tx);
          py := py + vars[1] * cos(ty);
        end;
        if varsi[2] > 0 then begin // complex
          r := tx * tx + ty * ty + 1E-6;
          px := px + vars[2] * tx/r;
          py := py + vars[2] * ty/r;
        end;
        if varsi[3] > 0 then begin // swirl
          r := tx * tx + ty * ty;
          px := px + vars[3] * (sin(r) * tx - cos(r) * ty);
          py := py + vars[3] * (cos(r) * tx + sin(r) * ty);
        end;
        if varsi[4] > 0 then begin // swirl
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;

          px := px + vars[4] * (sin(a) * tx - cos(a) * ty);
          py := py + vars[4] * (cos(a) * tx + sin(a) * ty);
        end;
        if varsi[5] > 0 then begin // no name
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)/ PI
          else
            a := 0;

          r := sqrt(tx * tx + ty * ty) - 1;

          px := px + vars[5] * a;
          py := py + vars[5] * r;
        end;

        if varsi[6] > 0 then begin // bent
          nx := tx;
          ny := ty;
          if nx < 0 then  nx := nx * 2;
          if ny < 0 then  ny := ny / 2;

          px := px + vars[6] * nx;
          py := py + vars[6] * ny;
        end;

        if varsi[7] > 0 then begin // Hart shaped box
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[7] * (sin(a * r) * r);
          py := py - vars[7] * (cos(a * r) * r);
        end;

        if varsi[8] > 0 then begin // The world in a sphere
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty) ;

          px := px + vars[8] * (sin(r) * (a));
          py := py + vars[8] * (cos(r) * (a));
        end;

        if varsi[9] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := power(tx * tx + ty * ty, 0.5) + 1E-6;

          px := px + vars[9] * ((cos(a) + sin(r)) / r);
          py := py + vars[9] * ((sin(a) - cos(r)) / r);
        end;

        if varsi[10] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := power(tx * tx + ty * ty, 0.25) + 1E-6;

          px := px + vars[10] * (sin(a) /r);
          py := py - vars[10] * (cos(a) * r);
        end;

        if varsi[11] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty) + 1E-6;

          px := px + vars[11] * ( sin(a) /r);
          py := py - vars[11] * ( cos(a) /r);
        end;

        if varsi[12] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[12] * ( sin(a ) * cos(r) );
          py := py - vars[12] * ( cos(a ) * sin(r));
        end;

        if varsi[13] > 0 then begin // Test
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[13] * (sign(sin(a + r)) * sqr(sin(a + r)) * r);
          py := py - vars[13] * (sign(cos(a - r)) * sqr(cos(a - r)) * r);
        end;

        if varsi[14] > 0 then begin // Square
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[14] * (sign(sin(a + r)) * sqr(sin(a + r)) * r);
          py := py - vars[14] * (sign(cos(a + r)) * sqr(cos(a + r)) * r);
        end;

        if varsi[15] > 0 then begin // Folded hankercief
          if (tx < -1e-10) or (tx > 1e-10) or (ty < -1e-10) or (ty > 1e-10) then
            a := arctan2(tx,ty)
          else
            a := 0;
          r := sqrt(tx * tx + ty * ty);

          px := px + vars[15] * (sin(a + r) * r);
          py := py - vars[15] * (cos(a - r) * r);
        end;
      except
        on EMathError do begin
//          raise Exception.Create('Iteration blows up');
          exit;
        end;
      end;

//      pix := trunc(px * 1024);
//      piy := trunc(py * 1024);
    end;

    // store points
    if i >= 0 then begin
      Points[i][0] := trunc(px * 1024);
      Points[i][1] := trunc(py * 1024);
      Points[i][2] := pic div 1024;
    end
  end;
end;

procedure TControlPoint.PrepareForinteger;
var
  i,j: Integer;
begin
  PreparePropTable;

  for i:= 0 to NXFORMS - 1 do begin
    xforms[i].colori := round(256 * 1024 * xforms[i].color);
    for j := 0 to NVARS - 1 do begin
      xforms[i].Varsi[j] := Round(xforms[i].Vars[j] * 1024);
    end;

    xforms[i].varType := -1;
    for j := 0 to NVARS - 1 do begin
      if xforms[i].Varsi[j] = 0 then
        continue;

      if xforms[i].vartype < 0 then begin
        xforms[i].Vartype := j;
      end else begin
        xforms[i].varType := -1;
        break;
      end;
    end;

    xforms[i].ci00 := Round(xforms[i].c[0,0] * 1024);
    xforms[i].ci01 := Round(xforms[i].c[0,1] * 1024);
    xforms[i].ci10 := Round(xforms[i].c[1,0] * 1024);
    xforms[i].ci11 := Round(xforms[i].c[1,1] * 1024);
    xforms[i].ci20 := Round(xforms[i].c[2,0] * 1024 * 1024);
    xforms[i].ci21 := Round(xforms[i].c[2,1] * 1024 * 1024);
  end;

  formvartype := -1;
  for i := 0 to NXFORMS - 1 do begin
    if xforms[i].density = 0 then
      break;

    if formvartype < 0 then begin
      formvartype := xforms[i].vartype;
    end else begin
      if formvartype <> xforms[i].vartype then begin
        formvartype := -1;
        break;
      end;
    end;
  end;
end;



function TControlPoint.Clone: TControlPoint;
var
  sl: TStringList;
begin
  sl := TStringList.Create;
  SaveToStringlist(sl);
  Result := TControlPoint.Create;
  Result.ParseStringlist(sl);
  sl.Free;
end;

procedure TControlPoint.ParseStringList(sl: TStringlist);
var
  s: String;
  i: integer;
begin
  for i := 0 to sl.Count - 1 do begin
    s := s + sl[i] + ' ';
  end;
  ParseString(s);
end;

{ TXForm }

constructor TXForm.Create;
var
  i: Integer;
begin
  density := 0;
  Color := 0;
  Vars[0] := 1;
  for i := 1 to NVARS - 1 do begin
    Vars[i] := 0;
  end;
  c[0,0] := 1;
  c[0,1] := 0;
  c[1,0] := 0;
  c[1,1] := 1;
  c[2,0] := 0;
  c[2,1] := 0;
end;

procedure CreateSinTable;
begin
end;

initialization
  CreateSinTable;

end.
